#!/usr/bin/env php
<?php
/***********************************************************************
 *  SETTINGS
 * ********************************************************************/
define("VERBOSE", true);
define("MOVING_THRESHOLD",.04); //.05 = 50 metres
define("OPENGTS",false);
/* HOST                                                               */
define("IP_ADDR","0");	// "0" = listen all ip
define("TCP_PORT",5050);
/* DATABASE                                                           */
define("DBHOST","localhost");
define("DBUSER","dbuser");
define("DBPASS","dbpassword");
define("DBNAME","gpsd");
/* TRACKER                                                            */
define("POLL_TIME",60); // SET POLL TIME 20,30,60,300,600 default:60 secs
define("SPEED_CONV",1.852); //From NM mile to Km
define("DFLT_MSG","tracker");
/* SOCKET                                                             */
define("SOCK_RCV_TIMEOUT",120); // Socket receive timeout in seconds
/***********************************************************************
 * END SETTINGS
 * ********************************************************************/
// Do not edit here-----------------------------------------------------
error_reporting(E_ALL);
// Do not exit while waiting for connect...
set_time_limit(0);
// Turn implicit flush on
ob_implicit_flush();

$dblink = dbConnect();
$address =  IP_ADDR;
$port = TCP_PORT;
$allowedIMEI = getAllowedImei($dblink);
$sendPollTime = false;

// Create the socket and bind it to the host and port, with infinite loop.

if (($sock = socket_create(AF_INET, SOCK_STREAM, SOL_TCP)) < 0) {
   writeLog("socket_create() failed: ".socket_strerror($sock),true);
}
if (($ret = socket_bind($sock, $address, $port)) < 0) {
   writeLog("socket_bind() failed: ".socket_strerror($ret),true);
}
if (($ret = socket_listen($sock, 5)) < 0) {
   writeLog("socket_listen() failed: ".socket_strerror($ret),true);
}
do {
	if (!mysql_ping ($dblink)) {
		mysql_close($dblink);
		$dblink = dbConnect();
	}
	try {
		if (FALSE === ($msgsock = socket_accept($sock))) {
		  throw new Exception("socket_accept() failed: " . socket_strerror(socket_last_error($msgsock)));
		}
		socket_set_option($msgsock, SOL_SOCKET, SO_RCVTIMEO, array('sec' => SOCK_RCV_TIMEOUT,'usec' => 0));
		writeLog("CONNECT");
		if (FALSE === ($buf = socket_read($msgsock, 2048))) {
		  throw new Exception("socket_read() failed: " . socket_strerror(socket_last_error($msgsock)));
		}
		writeLog("RECEIVED: ".$buf);
		$actualIMEI = "";
		$outData = array();
		$buf = trim($buf); // clean up input string
		$dirtyMode = (substr($buf, 0, 2) == "##") ? false : true;
		$actualIMEI = (!$dirtyMode) ? substr($buf, 8, 15) : substr($buf, 5, 15);  // returns IMEI

		if (!in_array($actualIMEI, $allowedIMEI)){
		  throw new Exception("Received: $actualIMEI from $buf, IMEI not allowed");
		}
		if (!$dirtyMode){
			$output = "LOAD". "\n";
			writeLog("SEND: LOAD");
			// Send intructions
			socket_write($msgsock, $output, strlen($output));
			if (FALSE === ($buf = socket_read($msgsock, 2048))) {
		  		throw new Exception("socket_read() failed: " . socket_strerror(socket_last_error($msgsock)));
			}
			$buf = trim($buf);
			writeLog("RECEIVED: ".$buf);
			if (empty($buf)){
				 throw new Exception("Received: nothing, disconnect");
			}

			if (($sendPollTime === false)) {
				$output = "ON". "\n";
				writeLog("SEND: ON");
				socket_write($msgsock, $output, strlen($output));
				$output = "**,imei:".$actualIMEI."," . pollTimeString(POLL_TIME)."\n";
				socket_write($msgsock, $output, strlen($output));
				writeLog("SEND: ".$output);
				if (FALSE === ($buf = socket_read($msgsock, 2048))) {
					$sendPollTime = false;
					throw new Exception("socket_read() failed: " . socket_strerror(socket_last_error($msgsock)));
				}
				$buf = empty($buf) ? "NO DATA" : trim($buf);
				$sendPollTime = true;
			}
		}
		$outData = explode ( "," , $buf );
		writeLog("DATA: ".$buf);
		if(count($outData)>=5){
			$outDecodedData = decodeData($outData);
			if ($outDecodedData['DATA_FL'] == "F" || $outDecodedData['MSG'] !== DFLT_MSG){
				if (OPENGTS) {
					$res = updatePosOpenGTS($outDecodedData);
				}
				else{
					$res = updatePos($outDecodedData);
				}
			}
		}
	} catch (Exception $e) {
		writeLog(" ".$e->getMessage(),true);
	}
	socket_close($msgsock);
	writeLog("SOCKET CLOSE");
} while (true);
socket_close($sock);
dbClose($dblink);

function now(){
	return date("Y-m-d H:i:s", time());
}

function decodeData($arr){
/* *********************************************************************
*   0 = imei:000000000000000	[imei]
*   1 = tracker					[Msg: help me / low battery / stockade /
* 											dt /move / speed / tracker]
*   2 = 0809231929				[acquisition time: YYMMDDhhmm +8GMT cn]
*   3 = 13554900601				[adminphone?]
*   4 = F						[Data: F - full / L - low]
*   5 = 112909.397				[Time (HHMMSS.SSS)]
*   6 = A						[A = available?]
*   7 = 2234.4669				[Latitude (DDMM.MMMM)]
*   8 = N						[Lat direction: N / S]
*   9 = 11354.3287				[Longitude (DDDMM.MMMM)]
*  10 = E						[Lon direction: E / O]
*  11 = 0.11					[speed Mph]
***********************************************************************/
	$out = array();
	$out['IMEI'] = substr($arr[0], 5, 15);
	$out['MSG'] = trim($arr[1]);
	$out['ACQUISITION_TIME'] = substr($arr[2], 0, 2)."-".
						substr($arr[2], 2, 2)."-".substr($arr[2], 4, 2).
						" ".substr($arr[2],6,2).":".substr($arr[2],8,2);
	$out['ADMINPHONE'] = trim($arr[3]);
	$out['DATA_FL'] = trim($arr[4]);
	 if ($out['DATA_FL'] === "F"){
		 $out['TIME'] = substr($arr[5], 0, 2).":" . substr($arr[5], 2, 2).":" . sprintf("%2d",round(floatval(substr($arr[5], 4, 6))));
		 $out['AVAILABLE'] = $arr[6]==="A" ? 1 : 0;
		 $out['LAT'] = floatval(substr($arr[7], 0, 2)) +
					   floatval(substr($arr[7], 2, 7)) / 60;
		 $out['LAT'] = $arr[8]==="N" ? $out['LAT'] : -$out['LAT'];
		 $out['LON'] = floatval(substr($arr[9], 0, 3))  +
					   floatval(substr($arr[9], 3, 7)) / 60;
		 $out['LON'] = $arr[10]==="E" ? $out['LON'] : -$out['LON'];
		 $out['SPEED'] = floatval($arr[11]) * SPEED_CONV;
	}
	else {
		 $out['TIME'] = "00:00:00";
		 $out['AVAILABLE'] = 0;
		 $out['LAT'] = (float) 0;
		 $out['LON'] = (float) 0;
		 $out['SPEED'] = (float) 0;
	}
	return $out;
}

function dbConnect() {
	$dblink = mysql_connect(DBHOST, DBUSER, DBPASS);
	if (!$dblink) {
		writeLog("Could not connect: ".mysql_error(),true);
	}
	else {
		$db = mysql_select_db(DBNAME, $dblink);
		if (!$db) {
			writeLog("Can't use ".DBNAME.": ".mysql_error(),true);
		}
	}
	return $dblink;
}
function dbClose(){
	global $dblink;
	return mysql_close($dblink);
}
function getAllowedImei() {
	global $dblink;
	$arr = array();
	if ($dblink) {
		$query = OPENGTS ? "SELECT DISTINCT imeiNumber FROM Device WHERE imeiNumber <> ''" :
							"SELECT DISTINCT imeiNumber FROM devices";
		$res = mysql_query($query,$dblink);
		while ($row = mysql_fetch_array($res, MYSQL_NUM)) {
			$arr[] = $row[0];
		}
	}
	return $arr;
}
function pollTimeString($secs){
	$secs = intval($secs);
	switch ($secs) {
		case 20:
			$res = "20s";
			break;
		case 30:
			$res = "30s";
			break;
		case 60:
			$res = "01m";
			break;
		case 300:
			$res = "05m";
			break;
		case 600:
			$res = "10m";
			break;
		default:
			$res = "01m";
	}
	return "C,".$res;
}
function updatePos($data) {
	global $dblink;
	$res = false;
	try {
		$query = "SELECT id FROM devices WHERE imeiNumber = '".
				 $data['IMEI']."'";
		$res = @mysql_query($query,$dblink);
		$row = @mysql_fetch_row($res, MYSQL_NUM);
		if (!empty($row[0])){
			if (MOVING_THRESHOLD > 0) {
				$dst = getLastPointDistance($row[0],$data['LAT'],$data['LON'],$data['IMEI']);
				if ($dst <= MOVING_THRESHOLD && $data['MSG'] == DFLT_MSG) {
					writeLog("Point is inside moving threshold: ".$dst);
					return $res;
				}
			}
			$date = new DateTime();
			$timestamp = $date->getTimestamp();
			$acq_time = is_valid_date($data['ACQUISITION_TIME']) ? $data['ACQUISITION_TIME'] : now();
			$query = "INSERT INTO positions (msg,timestamp,acq_time,
			track_time,is_valid,latitude,longitude,speedKPH,
			course,device_id) VALUES ('".$data['MSG']."',".$timestamp.",'".
			$acq_time."','".$data['TIME']."',".
			$data['AVAILABLE'].",".$data['LAT'].",".$data['LON'].",".
			$data['SPEED'].",0,".$row[0].")";
			$res = @mysql_query($query,$dblink);
			if (mysql_errno($dblink)) {
				$msg = "MySQL error ".mysql_errno($dblink).": ".
						mysql_error($dblink)." Query= ".$query;
				writeLog($msg,true);
			}
		}
	} catch (Exception $e) {
		writeLog(" ".$e->getMessage(),true);
	}
	return $res;
}
function updatePosOpenGTS($data){
	global $dblink;
	$res = false;
	try {
		$query = "SELECT accountID, deviceID FROM Device WHERE imeiNumber = '".$data['IMEI']."'";
		$row = @mysql_fetch_row(mysql_query($query,$dblink),MYSQL_NUM);
		if (!empty($row[0])){
			if (MOVING_THRESHOLD > 0) {
				$dst = getLastPointDistance($row[0],$data['LAT'],$data['LON'],$data['IMEI']);
				if ($dst <= MOVING_THRESHOLD && $data['MSG'] == DFLT_MSG) {
					writeLog("Point is inside moving threshold: ".$dst);
					return $res;
				}
			}
			$date = new DateTime();
			$timestamp = $date->getTimestamp();
			$date = new DateTime($data['ACQUISITION_TIME']);
			$data['ACQUISITION_TIME'] = $date->format('Y-m-d H:i:s');
			$exactDate = mergeDateAndTime($data['ACQUISITION_TIME'],$data['TIME']);
			$acq_time = is_valid_date($exactDate) ? $exactDate : now();
			$rawData = implode(",",$data);
			$query = "INSERT INTO EventData VALUES ('".$row[0]."', '".$row[1].
			"', UNIX_TIMESTAMP('$acq_time'), ".$data['AVAILABLE'].", ".$data['LAT'].", ".$data['LON'].
			", 0, ".$data['SPEED'].", 0, 0, '', 0, '', '', '".$rawData."', 0, 0, 0, '0000000001', $timestamp)";
			$res = @mysql_query($query,$dblink);
			if (mysql_errno($dblink)) {
				throw new Exception("MySQL error ".mysql_errno($dblink).": ".mysql_error($dblink)." Query= ".$query);
			}
			if ($data['DATA_FL'] === "F"){
				$query = "UPDATE Device SET lastValidLatitude = ".$data['LAT'].", lastValidLongitude = ".$data['LON'].", lastGPSTimestamp = UNIX_TIMESTAMP('$acq_time'), lastUpdateTime = $timestamp WHERE imeiNumber = ".$data['IMEI'];
				$res = @mysql_query($query,$dblink);
				if (mysql_errno($dblink)) {
					throw new Exception("MySQL error ".mysql_errno($dblink).": ".mysql_error($dblink)." Query= ".$query);
				}
			}
		}
	} catch (Exception $e) {
		writeLog(" ".$e->getMessage(),true);
	}

}
function getLastPointDistance($id,$lat,$lon,$imei) {
	global $dblink;

	if (OPENGTS) {
		$query = "SELECT (6371 * acos(cos(radians(lastValidLatitude)) *
		cos(radians($lat)) * cos(radians($lon) - radians(lastValidLongitude))
		+ sin(radians(lastValidLatitude)) * sin(radians($lat)))) AS distance
		FROM Device WHERE imeiNumber = $imei
		ORDER BY lastUpdateTime DESC LIMIT 1";
	}
	else{
		$query = "SELECT (6371 * acos(cos(radians(latitude)) *
		cos(radians($lat)) * cos(radians($lon) - radians(longitude))
		+ sin(radians(latitude)) * sin(radians($lat)))) AS distance
		FROM positions WHERE device_id = $id AND is_valid
		ORDER BY timestamp DESC LIMIT 1";
	}
	$res = mysql_query($query,$dblink);
	$row = mysql_fetch_row($res, MYSQL_NUM);
	return abs($row[0]);
}
function is_valid_date($str){
	$str = str_replace("-"," ",$str);
	$str = str_replace(":"," ",$str);
	$arrDate = explode(" ",$str);
	foreach($arrDate as $val){
		if (!is_numeric($val)){
			return false;
		}
	}
	$res = checkdate(intval($arrDate[1]),intval($arrDate[2]),intval($arrDate[0]));
	$res = $res && ($arrDate[3] >=0 && $arrDate[3] <=24);
	$res = $res && ($arrDate[4] >=0 && $arrDate[4] <=59);
	return $res;
}
function mergeDateAndTime ($strDate,$strTime){
	$datePart = substr($strDate,0,8);
	return $datePart." ".$strTime;
}
function writeLog($msg,$is_error=false) {
	$msg = now()." ".$msg."\n";
	if($is_error){
		echo $msg;
	}
	else {
		if(VERBOSE)
			echo $msg;
	}
	return true;
}
?>


